# pate-syft

This repo contains the code to apply a variation of PATE model with DP in [pysyft](https://github.com/OpenMined/PySyft) 

Private Aggregation of Teacher Ensembles (PATE) is a technique that allows the training of machine learning models with guarantees through Differential Privacy. The technique proposes to train multiple “teachers” models on disjoint sets of sensitive private data, and then use these teachers to guide the training of a “student” model on unlabeled public data. In the training of the student a set of data is sent through each model of the teacher to obtain a prediction. The predictions of the teachers are aggregated to create an ensemble of teachers and return this result to the student.

This model might be useful for organizations to share models and data without actually compromise the information. 
This code focuses to provide a framework to resolve the following scenary:
![pate.png](pate.png)

Suppose that organizations O_1, O_2... O_n has data to create models. They can share their models and create this ensemble. And the organization Os can use that ensemble to train a model, without compromising the information about the data that each organization has as well as their models. 
In this case, we have the following components: 
- Another Organization O_s start a Student to get the information needed 
- TrustedCurator creates an ensemble from the results of the different models
- Teacher uses their trained model to make predictions, each model trained by the Organizations O_1, O_2...O_n 

## Create environment 
```bash
   python3.10  -m venv env      
```
## Activate environment 
```bash
    python3 -m venv env
    source env/bin/activate
```
## Install dependencies 
```bash
    pip install -r requirements.txt
    git pull https://github.com/OpenMined/PySyft.git
    git checkout dev
    git pull
    cd packages/hagrid/
    pip install -e .
```
*Observation:* The code works with python version 3.10.8 and syft==0.7.0b50

## [Launch domain with docker and hgrid](https://openmined.github.io/PySyft/deployment/index.html#local-deployment-using-docker) with the dev version 

```bash
    hagrid launch --dev --tail=false
```

## Clean domain  
```bash
    cd src 
    python clean.py 
```

## Setup (create users)
```python
   python setup.py  
```

## Delete containers
If you want to delete all the Docker containers in your machine execute the following lines.
> **Warning**
> Note that ALL containers will be removed 
```bash
    docker rm -f $(docker ps -a -q)  
    docker volume rm $(docker volume ls -q)
    docker network prune 
```

## Components 
You can find the code for each component at:
- [Student.py](src/Student.py): uploads a dataframe to a domain to be used by the teachers to predict, waits for the result from the ensemble of the teachers.   
- [Teacher.py](src/Teacher.py) : loads a model and calculates the predictions for the student's dataset, uploads the result to a domain
- [TrustedCurator.py](src/TrustedCurator.py): waits for the results of the teachers and performs the ensembl, uploads the result of the ensemble to a domain. If the ensemble is a classification problem, it computes the voting and if it is a prediction the average. 


# Start the components in order in different terminals:
## Classification problem 
### Start student
```bash
    source env/bin/activate
    cd src/
    python start_student_classifier.py
```

### Start the TrustedCuratorTeacher component, passing as parameter the amount of teachers, for example 2 teachers:
```bash
    source env/bin/activate
    cd src/
    python start_trusted_curator_binary_classifier.py 2
```

### Start the Teacher component, passing as parameter the identifier:
```bash
    source env/bin/activate
    cd src/
    python start_teacher_binary_classifier.py 0
```
```bash
    source env/bin/activate
    cd src/
    python start_teacher_binary_classifier.py 1
```

## Regression problem
### Start student
```bash
    source env/bin/activate
    cd src/
    python start_student_regression.py
```

### Start the TrustedCuratorTeacher component, passing as parameter the amount of teachers, for example 2 teachers:
```bash
    source env/bin/activate
    cd src/
    python start_trusted_curator_regression.py 2
```

### Start the Teacher component, passing as parameter the identifier:
```bash
    source env/bin/activate
    cd src/
    python start_teacher_regression.py 0
```
```bash
    source env/bin/activate
    cd src/
    python start_teacher_regression.py 1
```

# This are the steps followed 
1. student uploads data to domain1 (student's doman)
4. teachers download data from domain1 (student's doman)
5. teachers upload result data to domain2 and domain3 (teacher's domains) 
6. trusted curator download result data from domain2 and domain3 (teacher's domains) 
7. trusted curator aggregates data (using voting or average)
8. trusted curator uploads result to domain4 (trusted_curator's domain)
9. student downloads result from domain4 (trusted_curator's domain)


# How Pysyft garantees privacy?

**Create dataset**
It is necessary to set up 3 values:   
1. min_val
2. max_val
3. data_subjects  

min_val and max_val are the lowest and highest values in the dataset. 
For instance, if you have an image data the min_val would be 0 and the max value 255. 
Less privacy budget will be lost if the distance between min and max value is small. 
The Data Subjects are the people whose privacy we're protecting using Differential Privacy.
The min and max values are used to adjust the magnitude of noise added proportionally in the data. 

**Privacy budget**  
Each Data Scientist is asigned a differential privacy budget that goes down when data is accessed. Each of the Data Subjects (individuals) have their own privacy budget that is being set by the framework, when counting down the Data Scientist's budget these variables are taken into account in order to control individuals' privacy over the dataset. 

**Sigma**
The data scientists can access the data through pointers, and visualize the information. The data that this pointer contains can be access through *publish* method. This method uses the privacy budget approved by the Data Owner to access the data in a noised format that does not compromise the original dataset. The parameter *sigma* refers to the standard deviation for the privacy budget the Data Scientist plans to use.


**Observations**       
- The Teachers and the Trusted Curator now run only one time, the should be updated to be in a while loop to support multiple invocations and be able to make more predictions. In this case we have to take a look to the privacy budget, since it will reach a time that there will be an error since the privacy budget is not enough for getting the result. In this case, we can consider requesting more privacy budget. 
- TODO: The privacy budget used and the sigma need to be calculated correctly 
- TODO: teachers might register in the Trusted Curator and it will return an id 
- TODO: improve logging 
- TODO: create configuration
- TODO: create Docker container for each component

