import syft as sy
import pandas as pd
import utils
import time
import numpy as np


class TrustedCurator():
	"""
    Parameters:
      tc_student_domain - TrustedCurator's domian to expose data for Student 
      tc_student_user - user to connect to tc_student_domain as owner
      student_domain - Student's domain
      student_user - user to connect to the student_domain
      n_teachers - number of teachers
      dataset_name name of the dataset
      teachers_domains - list of teacher's domains and data_scientist users [(domian, data_scientist_user)]
    """
	def __init__(self, 
		tc_student_domain, 
		tc_student_user,
		student_domain, 
		student_user, 
		n_teachers, 
		dataset_name,
		min_max_vals={},
		teachers_domains=[],
		classes = 2,
		method=utils.Method.AVERAGE):
		self.tc_student_domain = tc_student_domain
		self.tc_student_user = tc_student_user
		self.student_domain = student_domain
		self.student_user = student_user
		self.n_teachers = n_teachers
		self.dataset_name = dataset_name
		self.teachers_domains = teachers_domains
		self.method = method 
		self.classes = classes
		self.min_max_vals = min_max_vals

	def get_pointers_average(self):
		pointers = []
		for t in range(self.n_teachers):
			print(f'getting result from teacher {t}')
			domain = utils.login(self.teachers_domains[t][0], self.teachers_domains[t][1])
			data_pointer = utils.get_pointer(domain, 
				f'{self.dataset_name}_teacher_{t}_result', 
				'result', retries=50)
			pointers.append(data_pointer)
		return pointers

	def get_pointers_voting(self):
			pointers = []
			for t in range(self.n_teachers):
				print(f'getting result from teacher {t}')
				domain = utils.login(self.teachers_domains[t][0], self.teachers_domains[t][1])
				pointers_class = []
				for index in range(0, self.classes):
					data_pointer = utils.get_pointer(domain, 
					f'{self.dataset_name}_teacher_{t}_result', 
					f'result_{index}', retries=50)
					pointers_class.append(data_pointer)
				pointers.append(pointers_class)
			return pointers

	def get_pointers(self):
		if self.method == utils.Method.AVERAGE:
			return self.get_pointers_average()
		else:
			return self.get_pointers_voting()


	def aggregate_average(self, pointers):
		result = pointers[0]
		for i in range(1, self.n_teachers):
		    result = result + pointers[i]
		result = result.publish(sigma=2)
		result = result.get()
		result = result/self.n_teachers
		df = pd.DataFrame()
		df['result'] = result
		return df

	def aggregate_voting(self, pointers):
		n_preds = pointers[0][0].shape[0]
		res_dfs = {}
		for c in range(0, self.classes):
		    res = pointers[0][c]
		    for i in range(1, self.n_teachers):
		        res = res + pointers[i][c]
		    res = res.publish(sigma=0.01)
		    res = res.get()
		    res_dfs[f'result_{c}'] = res    

		df = pd.DataFrame(res_dfs)
		df['result'] = df.idxmax(axis="columns")
		classes_values = {}
		for c in range(0, self.classes):
		    classes_values[f'result_{c}'] = c 
		df['result'] = df['result'].apply(lambda x:  classes_values[x])
		cols_to_drop = list(map(lambda x: f'result_{x}', list(range(0,self.classes))))
		df.drop(cols_to_drop, axis=1, inplace=True) 
		return df

	def aggregate(self, pointers): 
		if self.method==utils.Method.AVERAGE:
			return self.aggregate_average(pointers)
		else:
			return self.aggregate_voting(pointers)


	def run(self):
		try:
			print("Logging into domains")
			tc_student = utils.login(self.tc_student_domain, self.tc_student_user)
			student = utils.login(self.student_domain, self.student_user)

			print("Wait for teachers' results")
			pointers = self.get_pointers()

			print('Aggregate results from teachers')
			df = self.aggregate(pointers) #TODO check if are avaliable all teachers responses and if not , consider what to do
			print(df.head(10))

			print(f"Upload result {self.dataset_name}_tc_result")
			utils.upload_dataset(tc_student, f'{self.dataset_name}_tc_result', df, self.min_max_vals)

		except Exception as ex:
			print('error:',ex)


	